
__all__ = [
    "__title__",
    "__summary__",
    "__url__",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
]

__title__ = "escnn"
__summary__ = "E(n)-Equivariant CNNs Library for PyTorch"
__url__ = 'https://github.com/QUVA-Lab/escnn'
__version__ = "1.0.0"
__author__ = "Gabriele Cesa"
__email__ = "cesa.gabriele@gmail.com"
__license__ = "BSD 3-Clause Clear"
